# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab Remote client test base classes
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: skip

import abc

from guidata.qthelpers import get_std_icon, win32_fix_title_bar_background
from guidata.widgets.codeeditor import CodeEditor
from qtpy import QtWidgets as QW

from cdlclient.config import _


class HostWidget(QW.QWidget):
    """Host widget: menu with action buttons, log viewer"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.button_layout = QW.QVBoxLayout()
        self.logwidget = CodeEditor(self)
        self.logwidget.setMinimumWidth(500)
        grid_layout = QW.QGridLayout()
        grid_layout.addLayout(self.button_layout, 0, 0)
        grid_layout.addWidget(self.logwidget, 0, 1)
        self.setLayout(grid_layout)

    def log(self, message):
        """Log message"""
        self.logwidget.appendPlainText(message)

    def add_spacing(self, spacing: int) -> None:
        """Add spacing to button box"""
        self.button_layout.addSpacing(spacing)

    def add_label(self, text: str) -> None:
        """Add label to button box"""
        self.button_layout.addWidget(QW.QLabel(text))

    def add_widget(self, obj: QW.QWidget, spacing_before: int = 0) -> None:
        """Add widget (QWidget) to button box"""
        if spacing_before > 0:
            self.add_spacing(spacing_before)
        self.button_layout.addWidget(obj)

    def add_button(self, title, slot, spacing_before=0, icon=None):
        """Add button"""
        btn = QW.QPushButton(title)
        if icon is not None:
            btn.setIcon(get_std_icon(icon))
        btn.clicked.connect(lambda _checked=False: slot())
        self.add_widget(btn, spacing_before=spacing_before)
        return btn

    def add_stretch(self):
        """Add stretch to button box"""
        self.button_layout.addStretch()


class AbstractClientWindowMeta(type(QW.QMainWindow), abc.ABCMeta):
    """Mixed metaclass to avoid conflicts"""


class AbstractClientWindow(QW.QMainWindow, metaclass=AbstractClientWindowMeta):
    """Abstract client window, to embed DataLab or connect to it"""

    PURPOSE = None
    INIT_BUTTON_LABEL = None
    SIG_TITLES = ("Oscilloscope", "Digitizer", "Radiometer", "Voltmeter", "Sensor")
    IMA_TITLES = (
        "Camera",
        "Streak Camera",
        "Image Scanner",
        "Laser Beam Profiler",
        "Gated Imaging Camera",
    )

    def __init__(self):
        super().__init__()
        win32_fix_title_bar_background(self)
        self.setWindowTitle(_("Host application"))
        self.setWindowIcon(get_std_icon("ComputerIcon"))
        self.cdl = None  # CDLMainWindow instance
        self.host = HostWidget(self)
        self.setCentralWidget(self.host)
        self.setup_window()
        self.host.add_stretch()
        self.index_sigtitle = -1
        self.index_imatitle = -1

    @property
    def sigtitle(self):
        """Return current signal title index"""
        self.index_sigtitle = idx = (self.index_sigtitle + 1) % len(self.SIG_TITLES)
        return self.SIG_TITLES[idx]

    @property
    def imatitle(self):
        """Return current image title index"""
        self.index_imatitle = idx = (self.index_imatitle + 1) % len(self.IMA_TITLES)
        return self.IMA_TITLES[idx]

    def setup_window(self):
        """Setup window"""
        self.host.add_label(self.PURPOSE)
        add_btn = self.host.add_button
        add_btn(self.INIT_BUTTON_LABEL, self.init_cdl, 10, "DialogApplyButton")
        add_btn(_("Raise window"), self.raise_cdl, 0, "FileDialogToParent")
        self.add_additional_buttons()
        add_btn(_("Add signal objects"), self.add_signals, 10, "CommandLink")
        add_btn(_("Add image objects"), self.add_images, 0, "CommandLink")
        add_btn(_("Remove all objects"), self.remove_all, 5, "MessageBoxWarning")
        add_btn(_("Close DataLab"), self.close_cdl, 10, "DialogCloseButton")

    def add_additional_buttons(self):
        """Add additional buttons"""

    @abc.abstractmethod
    def init_cdl(self):
        """Open DataLab test"""

    def raise_cdl(self):
        """Raise DataLab window"""
        if self.cdl is not None:
            self.cdl.raise_window()
            self.host.log("=> Raised DataLab window")

    @abc.abstractmethod
    def close_cdl(self):
        """Close DataLab window"""

    @abc.abstractmethod
    def add_signals(self):
        """Add signals to DataLab"""

    @abc.abstractmethod
    def add_images(self):
        """Add images to DataLab"""

    @abc.abstractmethod
    def remove_all(self):
        """Remove all objects from DataLab"""



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab Remote client connection dialog example
"""

# guitest: show

from guidata.qthelpers import qt_app_context
from qtpy import QtWidgets as QW

from cdlclient import SimpleRemoteProxy
from cdlclient.widgets import ConnectionDialog


def test_dialog():
    """Test connection dialog"""
    proxy = SimpleRemoteProxy(autoconnect=False)
    with qt_app_context():
        dlg = ConnectionDialog(proxy.connect)
        if dlg.exec():
            QW.QMessageBox.information(None, "Connection", "Successfully connected")
        else:
            QW.QMessageBox.critical(None, "Connection", "Connection failed")


if __name__ == "__main__":
    test_dialog()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""Qt helpers"""

from __future__ import annotations

import os
import os.path as osp
from collections.abc import Generator
from contextlib import contextmanager

from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtSvg as QS
from qtpy import QtWidgets as QW

from cdlclient.config import MOD_PATH

WIDGETS_PATH = osp.join(MOD_PATH, "widgets")


def svgtext_to_icon(text: str) -> QG.QIcon:
    """Convert SVG text to QIcon

    Args:
        text: SVG text

    Returns:
        Icon
    """
    svg_bytes = QC.QByteArray(text.encode("utf-8"))
    renderer = QS.QSvgRenderer(svg_bytes)  # pylint: disable=no-member
    pixmap = QG.QPixmap(64, 64)  # You can adjust the size as needed
    pixmap.fill(QC.Qt.transparent)  # Fill the pixmap with transparency
    painter = QG.QPainter(pixmap)
    renderer.render(painter)
    painter.end()
    return QG.QIcon(pixmap)


def svgfile_to_base64(filename: str) -> bytes:
    """Convert SVG file to Base64-encoded bytes

    Args:
        filename: SVG filename

    Returns:
        Base64-encoded bytes
    """
    image = QG.QImage(filename)
    data = QC.QByteArray()
    buf = QC.QBuffer(data)
    image.save(buf, "PNG")
    return data.toBase64().data()


def imagefile_to_base64(filename: str) -> bytes:
    """Convert image file to Base64-encoded bytes

    Args:
        filename: image filename

    Returns:
        Base64-encoded bytes
    """
    image = QG.QImage(filename)
    data = QC.QByteArray()
    buf = QC.QBuffer(data)
    image.save(buf, "PNG")
    return data.toBase64().data()


def imagefile_to_python_module(filename: str, destmod: str) -> None:
    """Convert image file to Python module

    Args:
        filename: image filename
        destmod: destination module name
    """
    data = imagefile_to_base64(filename)
    destmod_path = osp.join(WIDGETS_PATH, destmod + ".py")
    if osp.isfile(destmod_path):
        os.remove(destmod_path)
    with open(destmod_path, "wb") as fn:
        fn.write("# -*- coding: utf-8 -*-\n\n".encode("utf-8"))
        fn.write("# pylint: skip-file\n\n".encode("utf-8"))
        fn.write("DATA = b'".encode("utf-8"))
        fn.write(data)
        fn.write("'".encode("utf-8"))


@contextmanager
def block_signals(widget: QW.QWidget, enable: bool) -> Generator[None, None, None]:
    """Eventually block/unblock widget Qt signals before/after doing some things
    (enable: True if feature is enabled)"""
    if enable:
        widget.blockSignals(True)
    try:
        yield
    finally:
        if enable:
            widget.blockSignals(False)
            widget.blockSignals(False)
            widget.blockSignals(False)
            widget.blockSignals(False)



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab Simple Client
=====================

DataLab Simple Client (`cdlclient`) is a Python library providing a proxy to `DataLab`_
application through XML-RPC protocol.

.. _DataLab: https://datalab-platform.com/
"""

# pylint: disable=unused-import
from cdlclient.baseproxy import SimpleBaseProxy  # noqa: F401
from cdlclient.remote import SimpleRemoteProxy  # noqa: F401

__version__ = "0.10.1"
__required_server_version__ = "0.14.2"
__docurl__ = "https://cdlclient.readthedocs.io/en/latest/"
__homeurl__ = "https://github.com/DataLab-Platform/DataLabSimpleClient/"



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab Remote client get object dialog example
"""

# guitest: show

from guidata.qthelpers import qt_app_context

from cdlclient import SimpleRemoteProxy
from cdlclient.widgets import GetObjectDialog


def test_dialog():
    """Test connection dialog"""
    proxy = SimpleRemoteProxy()
    with qt_app_context():
        # 1. Select an image or signal object
        dlg = GetObjectDialog(None, proxy)
        if dlg.exec():
            obj = proxy.get_object(dlg.get_current_object_uuid())
            print(str(obj))
        # 2. Select a signal object only
        dlg = GetObjectDialog(None, proxy, panel="signal")
        if dlg.exec():
            obj = proxy.get_object(dlg.get_current_object_uuid())
            print(str(obj))
        # 3. Select an image object only
        dlg = GetObjectDialog(None, proxy, panel="image")
        if dlg.exec():
            obj = proxy.get_object(dlg.get_current_object_uuid())
            print(str(obj))


if __name__ == "__main__":
    test_dialog()



# -*- coding: utf-8 -*-
"""
DataLab Simple Client Widgets
-----------------------------

This module requires `guidata`_ and `qtpy`_ packages.

.. _guidata: https://pypi.org/project/guidata/
.. _qtpy: https://pypi.org/project/QtPy/

"""

# pylint: disable=unused-import
from cdlclient.widgets.connection import ConnectionDialog  # noqa: F401
from cdlclient.widgets.objectdialog import GetObjectDialog  # noqa: F401



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""Convert PNG image to Python code"""

# guitest: skip

import os
import os.path as osp

from guidata.qthelpers import qt_app_context
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

from cdlclient.config import MOD_NAME, MOD_PATH
from cdlclient.qthelpers import imagefile_to_python_module

RES_PATH = osp.join(MOD_PATH, os.pardir, "resources")


def test_conv(filename: str, destmod: str) -> None:
    """Test image to code conversion

    Args:
        filename: image filename
        destmod: destination module name
    """
    with qt_app_context(exec_loop=True):
        widget = QW.QWidget()
        vlayout = QW.QVBoxLayout()
        widget.setLayout(vlayout)
        label1 = QW.QLabel()
        label1.setPixmap(QG.QPixmap(filename))
        label2 = QW.QLabel()
        imagefile_to_python_module(filename, destmod)
        mod = __import__(f"{MOD_NAME}.widgets.{destmod}", fromlist=[destmod])
        pixmap = QG.QPixmap()
        pixmap.loadFromData(QC.QByteArray.fromBase64(mod.DATA))
        label2.setPixmap(pixmap)
        vlayout.addWidget(label1)
        vlayout.addWidget(label2)
        widget.show()


if __name__ == "__main__":
    test_conv(osp.join(RES_PATH, "DataLab-Banner-200.png"), "datalab_banner")



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab Simple Model
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

from typing import Any

import guidata.dataset as gds
from guidata.io import JSONReader

ROI_KEY = "_roi_"
ANN_KEY = "_ann_"


class BaseObj:
    """Base object"""

    # This is overriden in children classes with a gds.DictItem instance:
    metadata: dict[str, Any] = {}

    def __set_annotations(self, annotations: str | None) -> None:
        """Set object annotations (JSON string describing annotation plot items)

        Args:
            annotations (str | None): JSON string describing annotation plot items,
                or None to remove annotations
        """
        if annotations is None:
            if ANN_KEY in self.metadata:
                self.metadata.pop(ANN_KEY)
        else:
            self.metadata[ANN_KEY] = annotations

    def __get_annotations(self) -> str:
        """Get object annotations (JSON string describing annotation plot items)"""
        return self.metadata.get(ANN_KEY, "")

    annotations = property(__get_annotations, __set_annotations)

    def get_annotated_shapes(self):
        """Get annotated shapes"""
        from plotpy.io import load_items  # pylint: disable=import-outside-toplevel

        if self.annotations:
            return load_items(JSONReader(self.annotations))
        return []


class SignalObj(gds.DataSet, BaseObj):
    """Signal object (simplified version of DataLab's Signal object)"""

    uuid = gds.StringItem("UUID").set_prop("display", hide=True)

    _tabs = gds.BeginTabGroup("all")

    _datag = gds.BeginGroup("Data and metadata")
    title = gds.StringItem("Signal title", default="Untitled")
    xydata = gds.FloatArrayItem("Data", transpose=True, minmax="rows")
    metadata = gds.DictItem("Metadata", default={})
    _e_datag = gds.EndGroup("Data and metadata")

    _unitsg = gds.BeginGroup("Titles and units")
    title = gds.StringItem("Signal title", default="Untitled")
    _tabs_u = gds.BeginTabGroup("units")
    _unitsx = gds.BeginGroup("X-axis")
    xlabel = gds.StringItem("Title", default="")
    xunit = gds.StringItem("Unit", default="")
    _e_unitsx = gds.EndGroup("X-axis")
    _unitsy = gds.BeginGroup("Y-axis")
    ylabel = gds.StringItem("Title", default="")
    yunit = gds.StringItem("Unit", default="")
    _e_unitsy = gds.EndGroup("Y-axis")
    _e_tabs_u = gds.EndTabGroup("units")
    _e_unitsg = gds.EndGroup("Titles and units")

    _e_tabs = gds.EndTabGroup("all")


class ImageObj(gds.DataSet, BaseObj):
    """Image object (simplified version of DataLab's Image object)"""

    uuid = gds.StringItem("UUID").set_prop("display", hide=True)

    _tabs = gds.BeginTabGroup("all")

    _datag = gds.BeginGroup("Data")
    data = gds.FloatArrayItem("Data")
    metadata = gds.DictItem("Metadata", default={})
    _e_datag = gds.EndGroup("Data")

    _dxdyg = gds.BeginGroup("Origin / Pixel spacing")
    _origin = gds.BeginGroup("Origin")
    x0 = gds.FloatItem("X<sub>0</sub>", default=0.0)
    y0 = gds.FloatItem("Y<sub>0</sub>", default=0.0).set_pos(col=1)
    _e_origin = gds.EndGroup("Origin")
    _pixel_spacing = gds.BeginGroup("Pixel spacing")
    dx = gds.FloatItem("Δx", default=1.0, nonzero=True)
    dy = gds.FloatItem("Δy", default=1.0, nonzero=True).set_pos(col=1)
    _e_pixel_spacing = gds.EndGroup("Pixel spacing")
    _e_dxdyg = gds.EndGroup("Origin / Pixel spacing")

    _unitsg = gds.BeginGroup("Titles / Units")
    title = gds.StringItem("Image title", default="Untitled")
    _tabs_u = gds.BeginTabGroup("units")
    _unitsx = gds.BeginGroup("X-axis")
    xlabel = gds.StringItem("Title", default="")
    xunit = gds.StringItem("Unit", default="")
    _e_unitsx = gds.EndGroup("X-axis")
    _unitsy = gds.BeginGroup("Y-axis")
    ylabel = gds.StringItem("Title", default="")
    yunit = gds.StringItem("Unit", default="")
    _e_unitsy = gds.EndGroup("Y-axis")
    _unitsz = gds.BeginGroup("Z-axis")
    zlabel = gds.StringItem("Title", default="")
    zunit = gds.StringItem("Unit", default="")
    _e_unitsz = gds.EndGroup("Z-axis")
    _e_tabs_u = gds.EndTabGroup("units")
    _e_unitsg = gds.EndGroup("Titles / Units")

    _e_tabs = gds.EndTabGroup("all")



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Module providing a connection dialog box for the DataLab proxy client
"""

from __future__ import annotations

import time
from collections.abc import Callable

from guidata.qthelpers import (
    get_std_icon,
    qt_app_context,
    win32_fix_title_bar_background,
)
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

from cdlclient.config import _
from cdlclient.qthelpers import svgtext_to_icon
from cdlclient.widgets import datalab_banner, svg_icons


class ConnectionThread(QC.QThread):
    """DataLab Connection thread"""

    SIG_CONNECTION_OK = QC.Signal()
    SIG_CONNECTION_KO = QC.Signal()

    def __init__(self, connect_callback: Callable, parent: QC.QObject = None) -> None:
        super().__init__(parent)
        self.connect_callback = connect_callback

    def run(self) -> None:
        """Run thread"""
        try:
            self.connect_callback()
            self.SIG_CONNECTION_OK.emit()
        except ConnectionRefusedError:
            self.SIG_CONNECTION_KO.emit()


class ConnectionDialog(QW.QDialog):
    """DataLab Connection dialog

    Args:
        connect_callback: Callback function to connect to DataLab server
        parent: Parent widget. Defaults to None.
    """

    def __init__(self, connect_callback: Callable, parent: QW.QWidget = None) -> None:
        super().__init__(parent)
        win32_fix_title_bar_background(self)
        self.setWindowTitle(_("Connection to DataLab"))
        self.setWindowIcon(svgtext_to_icon(svg_icons.DATALAB))
        self.resize(300, 50)
        self.host_label = QW.QLabel()
        pixmap = QG.QPixmap()
        pixmap.loadFromData(QC.QByteArray.fromBase64(datalab_banner.DATA))
        self.host_label.setPixmap(pixmap)
        self.host_label.setAlignment(QC.Qt.AlignCenter)
        self.progress_bar = QW.QProgressBar()
        self.progress_bar.setRange(0, 0)
        status = QW.QWidget()
        status_layout = QW.QHBoxLayout()
        status.setLayout(status_layout)
        self.status_label = QW.QLabel(_("Waiting for connection..."))
        self.status_icon = QW.QLabel()
        status_layout.addWidget(self.status_icon)
        status_layout.addWidget(self.status_label)
        status_layout.setContentsMargins(0, 0, 0, 0)
        status_layout.addStretch()
        layout = QW.QVBoxLayout()
        layout.addWidget(self.host_label)
        layout.addWidget(self.progress_bar)
        layout.addWidget(status)
        self.setLayout(layout)
        self.thread = ConnectionThread(connect_callback)
        self.thread.SIG_CONNECTION_OK.connect(self.__on_connection_successful)
        self.thread.SIG_CONNECTION_KO.connect(self.__on_connection_failed)
        button_box = QW.QDialogButtonBox(QW.QDialogButtonBox.Cancel)
        button_box.rejected.connect(self.reject)
        layout.addWidget(button_box)

    def __set_status_icon(self, name: str) -> None:
        """Set status icon with standard Qt icon name"""
        self.status_icon.setPixmap(QG.QPixmap(get_std_icon(name).pixmap(24)))

    def __connect_to_server(self) -> None:
        """Connect to server"""
        self.progress_bar.setRange(0, 0)
        self.__set_status_icon("BrowserReload")
        self.status_label.setText(_("Connecting to server..."))
        self.thread.start()

    def __on_connection_successful(self) -> None:
        """Connection successful"""
        self.progress_bar.setRange(0, 1)
        self.progress_bar.setValue(1)
        self.__set_status_icon("DialogApplyButton")
        self.status_label.setText(_("Connection successful!"))
        QC.QTimer.singleShot(1000, self.accept)

    def __on_connection_failed(self) -> None:
        """Connection failed"""
        self.progress_bar.setRange(0, 1)
        self.progress_bar.setValue(1)
        self.__set_status_icon("MessageBoxCritical")
        self.status_label.setText(_("Connection failed."))
        QC.QTimer.singleShot(2000, self.reject)

    def exec(self) -> int:
        """Execute dialog"""
        self.__connect_to_server()
        return super().exec()


if __name__ == "__main__":

    def fake_connect():
        """Fake connection"""
        time.sleep(5)
        # raise ConnectionRefusedError("Connection refused")

    with qt_app_context():
        dlg = ConnectionDialog(fake_connect)
        dlg.exec()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab Simple Client configuration module
------------------------------------------

This module handles `DataLab Simple Client` configuration.
"""

from __future__ import annotations

from guidata import configtools

MOD_NAME = "cdlclient"
_ = configtools.get_translation(MOD_NAME)

MOD_PATH = configtools.get_module_data_path(MOD_NAME)



# -*- coding: utf-8 -*-

# pylint: skip-file

import os
import sys

sys.path.insert(0, os.path.abspath(".."))

import cdlclient  # noqa: E402

os.environ["CDL_DOC"] = "1"

# -- Project information -----------------------------------------------------

project = "DataLab Simple Client"
author = ""
copyright = "2023, Codra - Pierre Raybaut"
html_logo = "_static/DataLab-title.png"
latex_logo = "_static/DataLab-Frontpage.png"
release = cdlclient.__version__

# -- General configuration ---------------------------------------------------

extensions = ["sphinx.ext.intersphinx", "sphinx.ext.napoleon", "sphinx.ext.mathjax"]
templates_path = ["_templates"]
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

html_theme = "pydata_sphinx_theme"
html_theme_options = {"show_toc_level": 2}
html_static_path = ["_static"]

# -- Options for sphinx-intl package -----------------------------------------

locale_dirs = ["locale/"]  # path is example but recommended.
gettext_compact = False  # optional.

# -- Options for autodoc extension -------------------------------------------

autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
}

# -- Options for intersphinx extension ---------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "guidata": ("https://guidata.readthedocs.io/en/latest/", None),
}



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab Remote client unit test
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code
# guitest: skip

from __future__ import annotations

import os.path as osp
import tempfile
import time
from collections.abc import Generator
from contextlib import contextmanager

import numpy as np
from guidata.env import execenv
from plotpy.builder import make

from cdlclient.remote import SimpleRemoteProxy


@contextmanager
def temporary_directory() -> Generator[str, None, None]:
    """Create a temporary directory and clean-up afterwards"""
    tmp = tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
    try:
        yield tmp.name
    finally:
        try:
            tmp.cleanup()
        except (PermissionError, RecursionError):
            pass


def multiple_commands(remote: SimpleRemoteProxy):
    """Execute multiple XML-RPC commands"""
    with temporary_directory() as tmpdir:
        x = np.linspace(-10, 10, 1000)
        y = np.sin(x)
        remote.add_signal("tutu", x, y)

        z = np.random.rand(200, 200)
        remote.add_image("toto", z)
        rect = make.annotated_rectangle(100, 100, 200, 200, title="Test")
        area = rect.get_rect()
        remote.add_annotations_from_items([rect])
        uuid = remote.get_sel_object_uuids()[0]
        items = remote.get_object_shapes()
        assert len(items) == 1 and items[0].get_rect() == area
        execenv.print("OK")
        remote.add_label_with_title(f"Image uuid: {uuid}")
        remote.select_groups([1])
        remote.select_objects([uuid])
        remote.delete_metadata()

        fname = osp.join(tmpdir, osp.basename("remote_test.h5"))
        remote.save_to_h5_file(fname)
        remote.reset_all()
        remote.open_h5_files([fname], True, False)
        remote.import_h5_file(fname, True)
        remote.set_current_panel("signal")
        assert remote.get_current_panel() == "signal"
        remote.calc("log10")

        remote.calc("fft")

        time.sleep(2)  # Avoid permission error when trying to clean-up temporary files


def test():
    """Remote client test"""
    remote = SimpleRemoteProxy()
    execenv.print("Executing multiple commands...", end="")
    multiple_commands(remote)
    execenv.print("OK")


if __name__ == "__main__":
    test()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab Remote client application test
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code

# guitest: show

from __future__ import annotations

import functools
from contextlib import contextmanager

import numpy as np
from guidata.env import execenv
from guidata.qthelpers import qt_app_context, qt_wait
from qtpy import QtWidgets as QW

from cdlclient import SimpleRemoteProxy
from cdlclient.config import _
from cdlclient.tests.remoteclient_base import AbstractClientWindow
from cdlclient.tests.remoteclient_unit import multiple_commands
from cdlclient.widgets import ConnectionDialog, GetObjectDialog

APP_NAME = "Remote client test"


def try_send_command():
    """Try and send command to DataLab application remotely"""

    def try_send_command_decorator(func):
        """Try... except... decorator"""

        @functools.wraps(func)
        def method_wrapper(*args, **kwargs):
            """Decorator wrapper function"""
            self: HostWindow = args[0]  # extracting 'self' from method arguments
            output = None
            try:
                output = func(*args, **kwargs)
            except ConnectionRefusedError:
                self.cdl = None
                message = "🔥 Connection refused 🔥 (server is not ready?)"
                self.host.log(message)
                QW.QMessageBox.critical(self, APP_NAME, message)
            return output

        return method_wrapper

    return try_send_command_decorator


class HostWindow(AbstractClientWindow):
    """Test main view"""

    PURPOSE = _("This the client application, which connects to DataLab.")
    INIT_BUTTON_LABEL = _("Connect to DataLab")
    SIG_TITLES = ("Oscilloscope", "Digitizer", "Radiometer", "Voltmeter", "Sensor")
    IMA_TITLES = (
        "Camera",
        "Streak Camera",
        "Image Scanner",
        "Laser Beam Profiler",
        "Gated Imaging Camera",
    )

    def init_cdl(self):
        """Open DataLab test"""
        if self.cdl is None:
            self.cdl = SimpleRemoteProxy(autoconnect=False)
            connect_dlg = ConnectionDialog(self.cdl.connect, self)
            ok = connect_dlg.exec()
            if ok:
                self.host.log("✨ Initialized DataLab connection ✨")
                self.host.log(f"  Communication port: {self.cdl.port}")
                self.host.log("  List of exposed methods:")
                for name in self.cdl.get_method_list():
                    self.host.log(f"    {name}")
            else:
                self.cdl = None
                self.host.log("🔥 Connection refused 🔥 (server is not ready?)")

    @try_send_command()
    def close_cdl(self):
        """Close DataLab window"""
        if self.cdl is not None:
            self.cdl.close_application()
            self.host.log("🎬 Closed DataLab!")
            self.cdl = None

    def add_additional_buttons(self):
        """Add additional buttons"""
        add_btn = self.host.add_button
        add_btn(_("Execute multiple commands"), self.exec_multiple_cmd, 10)
        add_btn(_("Get object titles"), self.get_object_titles, 10)
        add_btn(_("Get object uuids"), self.get_object_uuids, 10)
        add_btn(_("Get object"), self.get_object)
        add_btn(_("Get object using dialog box"), self.get_object_dialog)

    @try_send_command()
    def exec_multiple_cmd(self):
        """Execute multiple commands in DataLab"""
        if self.cdl is not None:
            self.host.log("Starting command sequence...")
            multiple_commands(self.cdl)
            self.host.log("...end")

    @try_send_command()
    def get_object_titles(self):
        """Get object (signal/image) titles for current panel"""
        if self.cdl is not None:
            self.host.log("Object titles:")
            titles = self.cdl.get_object_titles()
            if titles:
                for name in titles:
                    self.host.log(f"  {name}")
            else:
                self.host.log("  Empty.")

    @try_send_command()
    def get_object_uuids(self):
        """Get object (signal/image) uuids for current panel"""
        if self.cdl is not None:
            self.host.log("Object uuids:")
            uuids = self.cdl.get_object_uuids()
            if uuids:
                for uuid in uuids:
                    self.host.log(f"  {uuid}")
            else:
                self.host.log("  Empty.")

    @try_send_command()
    def get_object(self):
        """Get object (signal/image) at index for current panel"""
        if self.cdl is not None:
            titles = self.cdl.get_object_titles()
            if titles:
                obj = self.cdl.get_object()
                self.host.log(f"Object '{obj.title}'")
                self.host.log(str(obj))
            else:
                self.host.log("🏴‍☠️ Object list is empty!")

    @try_send_command()
    def get_object_dialog(self):
        """Get object (signal/image) using dialog box"""
        if self.cdl is not None:
            dialog = GetObjectDialog(self, self.cdl)
            if dialog.exec():
                uuid = dialog.get_current_object_uuid()
                obj = self.cdl.get_object(uuid)
                self.host.log(f"Object '{obj.title}'")
                self.host.log(str(obj))

    def add_signals(self):
        """Add signals to DataLab"""
        if self.cdl is not None:
            x = np.linspace(0, 10, 1000)
            for title, y in (
                ("Sinus", np.sin(x)),
                ("Cosinus", np.cos(x)),
                ("Tan", np.tan(x)),
            ):
                self.cdl.add_signal(title, x, y)
                self.host.log(f"Added signal: {title}")

    def add_images(self):
        """Add images to DataLab"""
        if self.cdl is not None:
            for title, z in (
                ("Zeros", np.zeros((100, 100))),
                ("Ones", np.ones((100, 100))),
                ("Random", np.random.random((100, 100))),
            ):
                self.cdl.add_image(title, z)
                self.host.log(f"Added image: {title}")

    @try_send_command()
    def remove_all(self):
        """Remove all objects from DataLab"""
        if self.cdl is not None:
            self.cdl.reset_all()
            self.host.log("Removed all objects")


@contextmanager
def qt_wait_print(dt: float, message: str):
    """Wait and print message"""
    qt_wait(dt)
    execenv.print(f"{message}...", end="")
    yield
    execenv.print("OK")


def test_remote_client():
    """Remote client test"""
    with qt_app_context(exec_loop=True):
        window = HostWindow()
        window.resize(800, 800)
        window.show()
        dt = 1
        if execenv.unattended:
            qt_wait(2)
            window.init_cdl()
            with qt_wait_print(dt, "Executing multiple commands"):
                window.exec_multiple_cmd()
            with qt_wait_print(dt, "Raising DataLab window"):
                window.raise_cdl()
            with qt_wait_print(dt, "Getting object titles"):
                window.get_object_titles()
            with qt_wait_print(dt, "Getting object uuids"):
                window.get_object_uuids()
            with qt_wait_print(dt, "Getting object"):
                window.cdl.select_objects([1])
                window.get_object()
            with qt_wait_print(dt, "Adding signals"):
                window.add_signals()
            with qt_wait_print(dt, "Adding images"):
                window.add_images()
            with qt_wait_print(dt, "Removing all objects"):
                window.remove_all()
            with qt_wait_print(dt, "Closing DataLab"):
                window.close_cdl()


if __name__ == "__main__":
    test_remote_client()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab Simple Client unit tests
"""

from __future__ import annotations

from guidata.guitest import run_testlauncher

import cdlclient


def run() -> None:
    """Run DataLab test launcher"""
    run_testlauncher(cdlclient)


if __name__ == "__main__":
    run()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Object dialog
-------------

This module provides a dialog box to select an object (signal or image) from a list.

.. autoclass:: GetObjectDialog
    :members:
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

import os
from collections.abc import Iterator
from typing import TYPE_CHECKING

from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW

from cdlclient.config import _
from cdlclient.qthelpers import block_signals, svgtext_to_icon
from cdlclient.widgets import datalab_banner, svg_icons

if TYPE_CHECKING:  # pragma: no cover
    from cdlclient import SimpleRemoteProxy


class SimpleObjectTree(QW.QTreeWidget):
    """Base object handling panel list widget, object (sig/ima) lists"""

    SIG_ITEM_DOUBLECLICKED = QC.Signal(str)
    SIG_CONTEXT_MENU = QC.Signal(QC.QPoint)

    def __init__(self, parent: QW.QWidget) -> None:
        self.__obj_uuids: list[list[str]] = []
        self.__obj_titles: list[list[str]] = []
        self.__grp_titles: list[str] = []
        self.__panel_name: str = ""
        super().__init__(parent)
        self.setHeaderHidden(True)
        self.setColumnCount(1)
        self.setAlternatingRowColors(True)
        self.itemDoubleClicked.connect(self.item_double_clicked)

    def __str__(self) -> str:
        """Return string representation"""
        textlist = []
        for tl_index in range(self.topLevelItemCount()):
            tl_item = self.topLevelItem(tl_index)
            textlist.append(tl_item.text(0))
            for index in range(tl_item.childCount()):
                textlist.append("    " + tl_item.child(index).text(0))
        return os.linesep.join(textlist)

    def initialize(self, proxy: SimpleRemoteProxy) -> None:
        """Initialize tree with objects, using proxy"""
        grp_titles, obj_uuids, obj_titles = proxy.get_group_titles_with_object_infos()
        self.__grp_titles = grp_titles
        self.__obj_uuids = obj_uuids
        self.__obj_titles = obj_titles
        self.__panel_name = proxy.get_current_panel()
        self.populate_tree()

    def iter_items(
        self, item: QW.QTreeWidgetItem | None = None
    ) -> Iterator[QW.QTreeWidgetItem]:
        """Recursively iterate over all items"""
        if item is None:
            for index in range(self.topLevelItemCount()):
                yield from self.iter_items(self.topLevelItem(index))
        else:
            yield item
            for index in range(item.childCount()):
                yield from self.iter_items(item.child(index))

    def get_item_from_id(self, item_id) -> QW.QTreeWidgetItem:
        """Return QTreeWidgetItem from id (stored in item's data)"""
        for item in self.iter_items():
            if item.data(0, QC.Qt.UserRole) == item_id:
                return item
        return None

    def get_current_item_id(self, object_only: bool = False) -> str | None:
        """Return current item id"""
        item = self.currentItem()
        if item is not None and (not object_only or item.parent() is not None):
            return item.data(0, QC.Qt.UserRole)
        return None

    def set_current_item_id(self, uuid: str, extend: bool = False) -> None:
        """Set current item by id"""
        item = self.get_item_from_id(uuid)
        if extend:
            self.setCurrentItem(item, 0, QC.QItemSelectionModel.Select)
        else:
            self.setCurrentItem(item)

    @staticmethod
    def __update_item(
        item: QW.QTreeWidgetItem, number: int, prefix: str, title: str, uuid: str
    ) -> None:
        """Update item"""
        item.setText(0, f"{prefix}{number:03d}: {title}")
        item.setData(0, QC.Qt.UserRole, uuid)

    def populate_tree(self) -> None:
        """Populate tree with objects"""
        uuid = self.get_current_item_id()
        with block_signals(widget=self, enable=True):
            self.clear()
        for grp_idx, (grp_title, obj_uuids, obj_titles) in enumerate(
            zip(self.__grp_titles, self.__obj_uuids, self.__obj_titles)
        ):
            self.add_group_item(grp_idx + 1, grp_title, obj_uuids, obj_titles)
        if uuid is not None:
            self.set_current_item_id(uuid)

    def __add_to_group_item(
        self, number: int, obj_uuid: str, obj_title: str, group_item: QW.QTreeWidgetItem
    ) -> None:
        """Add object to group item"""
        item = QW.QTreeWidgetItem()
        prefix = self.__panel_name[0]
        svgtext = svg_icons.SIGNAL if prefix == "s" else svg_icons.IMAGE
        item.setIcon(0, svgtext_to_icon(svgtext))
        self.__update_item(item, number, prefix, obj_title, obj_uuid)
        group_item.addChild(item)

    def add_group_item(
        self, number: int, title: str, obj_uuids: list[str], obj_titles: list[str]
    ) -> None:
        """Add group item"""
        group_item = QW.QTreeWidgetItem()
        group_item.setIcon(0, svgtext_to_icon(svg_icons.GROUP))
        self.__update_item(group_item, number, "g", title, "")
        self.addTopLevelItem(group_item)
        group_item.setExpanded(True)
        for obj_idx, (obj_uuid, obj_title) in enumerate(zip(obj_uuids, obj_titles)):
            self.__add_to_group_item(obj_idx + 1, obj_uuid, obj_title, group_item)

    def item_double_clicked(self, item: QW.QTreeWidgetItem) -> None:
        """Item was double-clicked: open a pop-up plot dialog"""
        if item.parent() is not None:
            oid = item.data(0, QC.Qt.UserRole)
            self.SIG_ITEM_DOUBLECLICKED.emit(oid)

    def contextMenuEvent(self, event: QG.QContextMenuEvent) -> None:  # pylint: disable=C0103
        """Override Qt method"""
        self.SIG_CONTEXT_MENU.emit(event.globalPos())


class GetObjectDialog(QW.QDialog):
    """Get object dialog box

    Args:
        parent: Parent widget
        proxy: Remote proxy
        panel: Panel to retrieve objects from ('signal', 'image' or None for both)
        title: Dialog title
    """

    def __init__(
        self,
        parent: QW.QWidget,
        proxy: SimpleRemoteProxy,
        panel: str | None = None,
        title: str | None = None,
    ) -> None:
        super().__init__(parent)
        assert panel in (None, "signal", "image")
        self.__proxy = proxy
        self.__current_object_uuid: str | None = None
        self.setWindowTitle(_("Select object") if title is None else title)
        self.setWindowIcon(svgtext_to_icon(svg_icons.DATALAB))
        vlayout = QW.QVBoxLayout()
        self.setLayout(vlayout)

        logo_label = QW.QLabel()
        pixmap = QG.QPixmap()
        pixmap.loadFromData(QC.QByteArray.fromBase64(datalab_banner.DATA))
        logo_label.setPixmap(pixmap)
        logo_label.setAlignment(QC.Qt.AlignCenter)

        panelgroup = None
        if panel is None:
            panelgroup = QW.QWidget()
            panellayout = QW.QHBoxLayout()
            panellayout.setContentsMargins(0, 0, 0, 0)
            # panellayout.setAlignment(QC.Qt.AlignCenter)
            panelgroup.setLayout(panellayout)

            panelcombo = QW.QComboBox()
            panelcombo.addItem(svgtext_to_icon(svg_icons.SIGNAL), _("Signals"))
            panelcombo.addItem(svgtext_to_icon(svg_icons.IMAGE), _("Images"))
            if proxy.get_current_panel() == "image":
                panelcombo.setCurrentIndex(1)
            panelcombo.currentIndexChanged.connect(self.__change_panel)

            panellabel = QW.QLabel(_("Active panel:"))
            panellayout.addWidget(panellabel)
            panellayout.addWidget(panelcombo)
            panellayout.setStretch(1, 1)
        else:
            self.__proxy.set_current_panel(panel)

        self.tree = SimpleObjectTree(parent)
        self.tree.initialize(proxy)
        self.tree.SIG_ITEM_DOUBLECLICKED.connect(lambda oid: self.accept())
        self.tree.itemSelectionChanged.connect(self.__current_object_changed)

        vlayout.addWidget(logo_label)
        vlayout.addSpacing(10)
        if panelgroup is not None:
            vlayout.addWidget(panelgroup)
        vlayout.addWidget(self.tree)

        bbox = QW.QDialogButtonBox(QW.QDialogButtonBox.Ok | QW.QDialogButtonBox.Cancel)
        bbox.accepted.connect(self.accept)
        bbox.rejected.connect(self.reject)
        self.ok_btn = bbox.button(QW.QDialogButtonBox.Ok)
        vlayout.addSpacing(10)
        vlayout.addWidget(bbox)
        # Update OK button state:
        self.__current_object_changed()

    def __change_panel(self, index: int) -> None:
        """Change panel"""
        self.__proxy.set_current_panel("signal" if index == 0 else "image")
        self.tree.initialize(self.__proxy)
        self.__current_object_changed()

    def __current_object_changed(self) -> None:
        """Item selection has changed"""
        self.__current_object_uuid = self.tree.get_current_item_id()
        self.ok_btn.setEnabled(bool(self.__current_object_uuid))

    def get_current_object_uuid(self) -> str:
        """Return current object uuid"""
        return self.__current_object_uuid



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab remote control
----------------------

This module provides utilities to control DataLab from a Python script (e.g. with
Spyder) or from a Jupyter notebook.

The :class:`SimpleRemoteProxy` class provides the main interface
to DataLab XML-RPC server.
"""

from __future__ import annotations

import configparser as cp
import importlib
import json
import os
import os.path as osp
import sys
import time
import warnings
from io import BytesIO
from xmlrpc.client import Binary, ServerProxy

import guidata.dataset as gds
import numpy as np
from guidata.env import execenv
from guidata.io import JSONReader, JSONWriter
from guidata.userconfig import get_config_basedir

import cdlclient
from cdlclient import simplemodel
from cdlclient.baseproxy import SimpleBaseProxy
from cdlclient.simplemodel import ImageObj, SignalObj

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code


XMLRPCPORT_ENV = "CDL_XMLRPCPORT"


def get_xmlrpcport_from_env() -> int | None:
    """Get XML-RPC port number from environment variable."""
    try:
        return int(os.environ.get(XMLRPCPORT_ENV))
    except (TypeError, ValueError):
        return None


def array_to_rpcbinary(data: np.ndarray) -> Binary:
    """Convert NumPy array to XML-RPC Binary object, with shape and dtype.

    The array is converted to a binary string using NumPy's native binary
    format.

    Args:
        data: NumPy array to convert

    Returns:
        XML-RPC Binary object
    """
    dbytes = BytesIO()
    np.save(dbytes, data, allow_pickle=False)
    return Binary(dbytes.getvalue())


def rpcbinary_to_array(binary: Binary) -> np.ndarray:
    """Convert XML-RPC binary to NumPy array.

    Args:
        binary: XML-RPC Binary object

    Returns:
        NumPy array
    """
    dbytes = BytesIO(binary.data)
    return np.load(dbytes, allow_pickle=False)


def dataset_to_json(param: gds.DataSet) -> list[str]:
    """Convert guidata DataSet to JSON data.

    The JSON data is a list of three elements:

    - The first element is the module name of the DataSet class
    - The second element is the class name of the DataSet class
    - The third element is the JSON data of the DataSet instance

    Args:
        param: guidata DataSet to convert

    Returns:
        JSON data
    """
    writer = JSONWriter()
    param.serialize(writer)
    param_json = writer.get_json()
    klass = param.__class__
    return [klass.__module__, klass.__name__, param_json]


def json_to_dataset(param_data: list[str]) -> gds.DataSet:
    """Convert JSON data to guidata DataSet.

    Args:
        param_data: JSON data

    Returns:
        guidata DataSet
    """
    param_module, param_clsname, param_json = param_data
    # Replacing DataLab's model by DataLab Simple Client's model:
    param_module = param_module.replace("cdl.core.model.signal", simplemodel.__name__)
    param_module = param_module.replace("cdl.core.model.image", simplemodel.__name__)
    mod = importlib.__import__(param_module, fromlist=[param_clsname])
    klass = getattr(mod, param_clsname)
    param = klass()
    reader = JSONReader(param_json)
    param.deserialize(reader)
    return param


# === Python 2.7 client side:
#
# # See DataLab's doc/remotecontrol_py27.py for an almost complete Python 2.7
# # implementation of SimpleRemoteProxy class
#
# import io
# from xmlrpclib import ServerProxy, Binary
# import numpy as np
# def array_to_binary(data):
#     """Convert NumPy array to XML-RPC Binary object, with shape and dtype"""
#     dbytes = io.BytesIO()
#     np.save(dbytes, data, allow_pickle=False)
#     return Binary(dbytes.getvalue())
# s = ServerProxy("http://127.0.0.1:8000")
# data = np.array([[3, 4, 5], [7, 8, 0]], dtype=np.uint16)
# s.add_image("toto", array_to_binary(data))


def get_cdl_xmlrpc_port():
    """Return DataLab current XML-RPC port"""
    if sys.platform == "win32" and "HOME" in os.environ:
        os.environ.pop("HOME")  # Avoid getting old WinPython settings dir
    fname = osp.join(get_config_basedir(), ".DataLab", "DataLab.ini")
    ini = cp.ConfigParser()
    ini.read(fname)
    try:
        return ini.get("main", "rpc_server_port")
    except (cp.NoSectionError, cp.NoOptionError) as exc:
        raise ConnectionRefusedError("DataLab has not yet been executed") from exc


def items_to_json(items: list) -> str | None:
    """Convert plot items to JSON string.

    Args:
        items (list): list of plot items

    Returns:
        str: JSON string or None if items is empty
    """
    from plotpy.io import save_items  # pylint: disable=import-outside-toplevel

    if items:
        writer = JSONWriter(None)
        save_items(writer, items)
        return writer.get_json(indent=4)
    return None


def json_to_items(json_str: str | None) -> list:
    """Convert JSON string to plot items.

    Args:
        json_str (str): JSON string or None

    Returns:
        list: list of plot items
    """
    from plotpy.io import load_items  # pylint: disable=import-outside-toplevel

    items = []
    if json_str:
        try:
            for item in load_items(JSONReader(json_str)):
                items.append(item)
        except json.decoder.JSONDecodeError:
            pass
    return items


def is_version_at_least(version1: str, version2: str) -> bool:
    """
    Compare two version strings to check if the first version is at least
    equal to the second.

    Args:
        version1 (str): The first version string.
        version2 (str): The second version string.

    Returns:
        bool: True if version1 is greater than or equal to version2, False otherwise.

    .. note::

        Development, alpha, beta, and rc versions are considered to be equal
        to the corresponding release version.
    """
    # Split the version strings into parts
    parts1 = [part.strip() for part in version1.split(".")]
    parts2 = [part.strip() for part in version2.split(".")]

    for part1, part2 in zip(parts1, parts2):
        if part1.isdigit() and part2.isdigit():
            if int(part1) > int(part2):
                return True
            elif int(part1) < int(part2):
                return False
        elif part1 > part2:
            return True
        elif part1 < part2:
            return False

    return len(parts1) >= len(parts2)


class SimpleRemoteProxy(SimpleBaseProxy):
    """Object representing a proxy/client to DataLab XML-RPC server.
    This object is used to call DataLab functions from a Python script.

    This is a subset of DataLab's `RemoteClient` class, with only the methods
    that do not require DataLab object model to be implemented.

    Args:
        autoconnect: If True, automatically connect to DataLab XML-RPC server.
         Defaults to True.

    Raises:
        ConnectionRefusedError: DataLab is currently not running

    Examples:
        Here is a simple example of how to use SimpleRemoteProxy in a Python script
        or in a Jupyter notebook:

        >>> from cdlclient import SimpleRemoteProxy
        >>> proxy = SimpleRemoteProxy()  # autoconnect is on by default
        Connecting to DataLab XML-RPC server...OK (port: 28867)
        >>> proxy.get_version()
        '1.0.0'
        >>> proxy.add_signal("toto", np.array([1., 2., 3.]), np.array([4., 5., -1.]))
        True
        >>> proxy.get_object_titles()
        ['toto']
        >>> proxy["toto"]
        <cdlclient.model.signal.SignalObj at 0x7f7f1c0b4a90>
        >>> "toto" in proxy
        True
        >>> proxy[1]
        <cdlclient.model.signal.SignalObj at 0x7f7f1c0b4a90>
        >>> proxy[1].data
        array([1., 2., 3.])
    """

    def __init__(self, autoconnect: bool = True) -> None:
        super().__init__()
        self.port: str = None
        self._cdl: ServerProxy
        if autoconnect:
            self.connect()

    def __connect_to_server(self, port: str | None = None) -> None:
        """Connect to DataLab XML-RPC server.

        Args:
            port (str | None): XML-RPC port to connect to. If not specified,
                the port is automatically retrieved from DataLab configuration.

        Raises:
            ConnectionRefusedError: DataLab is currently not running
        """
        if port is None:
            port = get_xmlrpcport_from_env()
            if port is None:
                port = get_cdl_xmlrpc_port()
        self.port = port
        self._cdl = ServerProxy(f"http://127.0.0.1:{port}", allow_none=True)
        try:
            version = self.get_version()
        except ConnectionRefusedError as exc:
            raise ConnectionRefusedError("DataLab is currently not running") from exc
        # If DataLab version is not compatible with this client, show a warning using
        # standard `warnings` module:
        if not is_version_at_least(version, cdlclient.__required_server_version__):
            warnings.warn(
                f"DataLab version {version} is not fully compatible with "
                f"DataLab Simple Client version {cdlclient.__version__}.\n"
                f"Please upgrade DataLab to {cdlclient.__required_server_version__} "
                f"or higher."
            )

    def connect(
        self,
        port: str | None = None,
        timeout: float | None = None,
        retries: int | None = None,
    ) -> None:
        """Try to connect to DataLab XML-RPC server.

        Args:
            port (str | None): XML-RPC port to connect to. If not specified,
                the port is automatically retrieved from DataLab configuration.
            timeout (float | None): Timeout in seconds. Defaults to 5.0.
            retries (int | None): Number of retries. Defaults to 10.

        Raises:
            ConnectionRefusedError: Unable to connect to DataLab
            ValueError: Invalid timeout (must be >= 0.0)
            ValueError: Invalid number of retries (must be >= 1)
        """
        timeout = 5.0 if timeout is None else timeout
        retries = 10 if retries is None else retries
        if timeout < 0.0:
            raise ValueError("timeout must be >= 0.0")
        if retries < 1:
            raise ValueError("retries must be >= 1")
        execenv.print("Connecting to DataLab XML-RPC server...", end="")
        for _index in range(retries):
            try:
                self.__connect_to_server(port=port)
                break
            except ConnectionRefusedError:
                time.sleep(timeout / retries)
        else:
            execenv.print("KO")
            raise ConnectionRefusedError("Unable to connect to DataLab")
        execenv.print(f"OK (port: {self.port})")

    def disconnect(self) -> None:
        """Disconnect from DataLab XML-RPC server."""
        # This is not mandatory with XML-RPC, but if we change protocol in the
        # future, it may be useful to have a disconnect method.
        self._cdl = None

    def is_connected(self) -> bool:
        """Return True if connected to DataLab XML-RPC server."""
        if self._cdl is not None:
            try:
                self.get_version()
                return True
            except ConnectionRefusedError:
                self._cdl = None
        return False

    def get_method_list(self) -> list[str]:
        """Return list of available methods."""
        return self._cdl.system.listMethods()

    # === Following methods should match the register functions in XML-RPC server

    def add_signal(
        self,
        title: str,
        xdata: np.ndarray,
        ydata: np.ndarray,
        xunit: str | None = None,
        yunit: str | None = None,
        xlabel: str | None = None,
        ylabel: str | None = None,
    ) -> bool:  # pylint: disable=too-many-arguments
        """Add signal data to DataLab.

        Args:
            title (str): Signal title
            xdata (numpy.ndarray): X data
            ydata (numpy.ndarray): Y data
            xunit (str | None): X unit. Defaults to None.
            yunit (str | None): Y unit. Defaults to None.
            xlabel (str | None): X label. Defaults to None.
            ylabel (str | None): Y label. Defaults to None.

        Returns:
            bool: True if signal was added successfully, False otherwise

        Raises:
            ValueError: Invalid xdata dtype
            ValueError: Invalid ydata dtype
        """
        xbinary = array_to_rpcbinary(xdata)
        ybinary = array_to_rpcbinary(ydata)
        return self._cdl.add_signal(
            title, xbinary, ybinary, xunit, yunit, xlabel, ylabel
        )

    # pylint: disable=too-many-arguments
    def add_image(
        self,
        title: str,
        data: np.ndarray,
        xunit: str | None = None,
        yunit: str | None = None,
        zunit: str | None = None,
        xlabel: str | None = None,
        ylabel: str | None = None,
        zlabel: str | None = None,
    ) -> bool:
        """Add image data to DataLab.

        Args:
            title (str): Image title
            data (numpy.ndarray): Image data
            xunit (str | None): X unit. Defaults to None.
            yunit (str | None): Y unit. Defaults to None.
            zunit (str | None): Z unit. Defaults to None.
            xlabel (str | None): X label. Defaults to None.
            ylabel (str | None): Y label. Defaults to None.
            zlabel (str | None): Z label. Defaults to None.

        Returns:
            bool: True if image was added successfully, False otherwise

        Raises:
            ValueError: Invalid data dtype
        """
        zbinary = array_to_rpcbinary(data)
        return self._cdl.add_image(
            title, zbinary, xunit, yunit, zunit, xlabel, ylabel, zlabel
        )

    def calc(self, name: str, param: gds.DataSet | None = None) -> gds.DataSet:
        """Call compute function ``name`` in current panel's processor.

        Args:
            name (str): Compute function name
            param (guidata.dataset.DataSet | None): Compute function
             parameter. Defaults to None.

        Returns:
            guidata.dataset.DataSet: Compute function result
        """
        if param is None:
            return self._cdl.calc(name)
        return self._cdl.calc(name, dataset_to_json(param))

    def get_object(
        self,
        nb_id_title: int | str | None = None,
        panel: str | None = None,
    ) -> SignalObj | ImageObj:
        """Get object (signal/image) from index.

        Args:
            nb_id_title: Object number, or object id, or object title.
             Defaults to None (current object).
            panel: Panel name. Defaults to None (current panel).

        Returns:
            Object

        Raises:
            KeyError: if object not found
        """
        param_data = self._cdl.get_object(nb_id_title, panel)
        if param_data is None:
            return None
        return json_to_dataset(param_data)

    def get_object_shapes(
        self,
        nb_id_title: int | str | None = None,
        panel: str | None = None,
    ) -> list:
        """Get plot item shapes associated to object (signal/image).

        Args:
            nb_id_title: Object number, or object id, or object title.
             Defaults to None (current object).
            panel: Panel name. Defaults to None (current panel).

        Returns:
            List of plot item shapes
        """
        items_json = self._cdl.get_object_shapes(nb_id_title, panel)
        return json_to_items(items_json)

    def add_annotations_from_items(
        self, items: list, refresh_plot: bool = True, panel: str | None = None
    ) -> None:
        """Add object annotations (annotation plot items).

        .. note:: This method is only available if PlotPy is installed.

        Args:
            items (list): annotation plot items
            refresh_plot (bool | None): refresh plot. Defaults to True.
            panel (str | None): panel name (valid values: "signal", "image").
                If None, current panel is used.
        """
        try:
            items_json = items_to_json(items)
        except ImportError as exc:
            raise ImportError("PlotPy is not installed") from exc
        if items_json is not None:
            self._cdl.add_annotations_from_items(items_json, refresh_plot, panel)
